import os 
import tkinter as tk

from PIL import ImageTk, Image

path = os.path.dirname(__file__)

pstl_ico_path = os.path.join(path, "pstl.ico")
pstl_png_path = os.path.join(path, "pstl.png")
pstl_16_ico_path = os.path.join(path, "pstl-16.ico")
pstl_32_ico_path = os.path.join(path, "pstl-32.ico")
